
#pragma once

// Genode includes
#include <base/log.h>
#include <util/string.h>

// protobuf includes
#include <google/protobuf/message.h>
#include <google/protobuf/util/time_util.h>

// stdcxx includes
#include <vector>

// gapfruit includes
#include <tee_session/tee_session.h>

// generated includes
#include <tee.pb.h>


namespace Tee
{
	using Genode::Hex;
	using Genode::Output;
	using Genode::log;

	using google::protobuf::Timestamp;
	using google::protobuf::Duration;

	using gapfruit::protobuf::tee::Address;
	using gapfruit::protobuf::tee::CounterRequest;
	using gapfruit::protobuf::tee::ExecuteInitRequest;
	using gapfruit::protobuf::tee::ExecuteInitResponse;
	using gapfruit::protobuf::tee::ExecuteRequest;
	using gapfruit::protobuf::tee::ExecuteResponse;
	using gapfruit::protobuf::tee::Hash;
	using gapfruit::protobuf::tee::QuoteRequest;
	using gapfruit::protobuf::tee::Request;
	using gapfruit::protobuf::tee::Signature;
	using gapfruit::protobuf::tee::SignedRequest;

	using Packet_content_ro = const char*;
	using Packet_content_rw = char*;

	struct Hexdump;

	struct Deserialize_failed  : Genode::Exception { };
	struct Serialize_failed    : Genode::Exception { };
	struct Transmission_error  : Genode::Exception { };

	using Certificate        = std::string;
	using Fingerprint_string = std::string;
	using Fingerprints       = std::vector<Fingerprint_string>;
	using Hash_string        = std::string;
	using Nonce_data         = std::string;
	using Request_data       = std::string;
	using Response_data      = std::string;
	using Security_domain    = std::string;

	struct Counter_response_data;
	struct Execute_init_response_data;
	struct Execute_request_data;
	struct Execute_response_data;
	struct Hash_data;
	struct Quote_response_data;
	struct Request_info;
	struct Signature_data;

	enum {
		NONCE_SIZE = 64,
	};

	// output the content of a packet to the log.
	void dump_packet(Packet_content_ro, size_t);

	// Create a Timestamp initialized with the current system time.
	Timestamp timestamp_from_system_time();
}


struct Tee::Hash_data
{
	Hash::Algorithm algorithm;
	Hash_string     value;
};


struct Tee::Signature_data
{
	Hash_string    signature;
	Certificate    certificate;
};


struct Tee::Request_info
{
	Timestamp        timestamp;
	Session_id       session_id;
	Security_domain  security_domain;
};


struct Tee::Counter_response_data
{
	uint64_t        counter_value;
	Signature_data  attestation_signature;
};


struct Tee::Execute_init_response_data
{
	Nonce_data      nonce_value;
	Signature_data  attestation_signature;
};


struct Tee::Execute_request_data
{
	Nonce_data    execution_id;
	Duration      timeout;
	Request_data  in_data;
};


struct Tee::Execute_response_data
{
	Response_data    out_data;
	Signature_data   out_signature;
	Hash_data        input_hash;
	uint64_t         counter;
	Hash_data        exec_hash;
	Hash_data        platform_state;
	Result           result;
	Timestamp        timestamp;
	Nonce_data       execution_id;
};


struct Tee::Quote_response_data
{
	const Fingerprints    allowed_clients;
	const uint64_t        counter_value;
	const Hash_data       platform_state;
	const uint64_t        request_max_size;
	const uint64_t        response_max_size;
	const Certificate     attestation_cert_chain;
	const Certificate     output_cert_chain;
	const Signature_data  attestation_signature;
};


struct Tee::Hexdump
{
	Tee::Packet_content_ro ptr;
	const size_t           offset;

	Hexdump(Tee::Packet_content_ro ptr, unsigned long offset) : ptr(ptr), offset(offset) { }

	void print(Genode::Output &out) const
	{
		using Genode::print;
		unsigned long off = offset;
		unsigned long *_ptr = (unsigned long *)(ptr + off);

		for (int i = 0; i < 4; i++) {
			print(out, Hex((uint32_t)off, Hex::OMIT_PREFIX, Hex::PAD), ": ");

			int j;
			for (j = 0; j < 5; j++)
				print(out, Hex(_ptr[(i * 8) + j], Hex::OMIT_PREFIX, Hex::PAD), "  ");

			off += j * sizeof(unsigned long);
			print(out, "\n");
		}
	}
};
